#!/usr/bin/perl
### sfn-move.perl  -*- Perl -*-
## Move .rsync-backup files under sfn/ directories as appropriate.

### Ivan Shmakov, 2020

## To the extent possible under law, the author(s) have dedicated
## all copyright and related and neighboring rights to this software
## to the public domain worldwide.  This software is distributed
## without any warranty.

## You should have received a copy of the CC0 Public Domain Dedication
## along with this software.  If not, see
## <http://creativecommons.org/publicdomain/zero/1.0/>.

### History:

## 0.2  2020-07-16
##      New --[no-]dry-run, --[no-]debug and -h, --help options.
##      Retrieve all sfn. references (were: just href= ones), thus
##      allowing support for non-HTML documents.  Use Getopt::Long.
##      Remove all (?:\.rsync-)?backup and \.~[0-9]+~ components from
##      filenames when looking for the current version (was: one of each.)
##      Fixed reference handling (were: references not loaded, suffix
##      not extracted, test inverted.)
##      ($usage, $example): New variables.
##      ($progname): Removed (now unused) variable.

## 0.1  2020-02-20 05:30:10Z
##      (sfn.kW2KaaGGSfQyn2KTk8NvCmnKkvAjpmXoK7oyVoPVsdw.perl)
##      Initial revision.

### Code:
use common::sense;
use English qw (-no_match_vars);

require Data::Dump;
require Digest::SHA;
# require Encode;
require Getopt::Long;
require IO::File;
# require List::Util;
# require POSIX;

our ($usage, $example) = (qq {Usage:
    \$ ${0} [-n|--[no-]debug] [--[no-]dry-run] [--] FILE...
    \$ ${0} -h|--help
}, qq {Example:

\$ ${0} -n -- sfn-move.perl.~[34]~ 
Warning: nGlS5LS76YTAs8jEv00M7w3gfRpmW-PYLRb7YP2gOiI.perl: unreferenced (sfn-move.perl.~3~); ignored
I: kW2KaaGGSfQyn2KTk8NvCmnKkvAjpmXoK7oyVoPVsdw from devel/scripts/sfn-move.perl.~4~
\$ 
});

Getopt::Long::Configure (qw (gnu_getopt));

our ($debug_p, $dry_run_p)
    = (0, 0);

my %cmdline_options =  (q (debug!)      => \$debug_p,
                        q (n|dry-run!)  => \$dry_run_p,
                        q (h|help)
                        => sub { print ($usage, "\n", $example); exit (0); });

my $parsable_p
    = Getopt::Long::GetOptions (%cmdline_options)
    or die ("Cannot parse command line arguments");

unless (@ARGV > 0) {
    print STDERR ($usage);
    ## .
    exit (1);
}

our $dev_fd_prefix
    = ($ENV{"DEVFDPREFIX"} // "/dev/fd/");

sub dev_fd {
    my ($fn) = @_;
    our ($dev_fd_prefix);
    my $l
        = length ($dev_fd_prefix);
    ## .
    (substr ($fn, 0, $l) eq $dev_fd_prefix
     ? substr ($fn, $l) : undef);
}

sub open_file {
    my ($fn, $mode) = @_;
    die ()
        if (index ($mode, ">") > 0);
    ## .
    return \*STDIN
        if ($fn eq "-");

    $mode
        //= "<";
    my $fd
        = ("GLOB" eq ref ($fn) ? $fn->fileno () : dev_fd ($fn));
    ## .
    (defined ($fd)
     ? IO::File->new_from_fd ($fd, $mode)
     : IO::File->new ($fn, $mode));
}

my $d0
    = Digest::SHA->new (256)
    or die ();
my %xref;
my %seen;
foreach my $fn (@ARGV) {
    my $fh
        = open_file ($fn)
        or die ($fn, ": Cannot open file: ", $!);
    $fh->binmode ()
        or die ($fn, ": ", $!);

    ## NB: non-greedy (.*?) matches are used below
    my $g
        = ($fn =~ s {(?:^|/)(?:\.(?:rsync-)?backup/)|\.~[0-9]+~$}{}rg);
    unless (exists ($seen{$g})) {
        my $gh
            = open_file ($g);
        unless ($gh and $gh->binmode ()) {
            warn ("Warning: ", $g, ": ", $!, "; ignored");
            next;
        };

        my @refs;
        while (<$gh>) {
            push (@refs, m { \b sfn\. (\S{43}[^][[:blank:](){}]*) }gx);
        }
        $gh->close ();
        warn ("D: ", $g, ": references: ", join (", ", @refs)) if ($debug_p);
        @xref{(@refs)}
            = ();
        @seen{($g)}
            = ();
    };

    my $d
        = $d0->clone ()->reset ();
    $d->addfile ($fh)
        or die ();
    my $b64
        = $d->b64digest ();
    my $b64u
        = ($b64 =~ tr {+/} {-_}r);
    my ($s)
        = ($fn =~ m { .* (\.[^.~]+) }x);
    my $h1
        = $b64u . $s;
    my $h
        = (($g =~ s {(.*/).*}{$1sfn/}r) . $h1);
    unless (exists ($xref{$h1})) {
        warn ("Warning: ", $h, ": unreferenced (", $g, "); ignored");
        next;
    }
    if (-l $h || -e $h) {
        warn ("Warning: ", $h, ": already exists; ignored");
        next;
    }
    warn ("D: ", $h, " from ", $fn) if ($debug_p);
    if (! $dry_run_p) {
        rename ($fn, $h);
    } elsif (! $debug_p) {
        warn ("I: ", $h, " from ", $fn, "\n");
    }
}

### Emacs trailer
## Local variables:
## coding: us-ascii
## fill-column: 72
## indent-tabs-mode: nil
## ispell-local-dictionary: "american"
## End:
### sfn-move.perl ends here
